@@ -29,6 +29,145 @@ from deepcinac.utils.cinac_file_utils import CinacFileReader, read_cell_type_cat
 from deepcinac.utils.display import plot_hist_distribution
 from deepcinac.cinac_structures import create_cinac_recording_from_cinac_file_segment
 import sklearn.metrics
+from collections import Counter
+
+
+def fusion_cell_type_predictions_by_type(cell_type_preds_dict, default_cell_type_pred,
+                                         cell_type_to_skip_if_conflict=None,
+                                         cell_type_config_file=None, filename_to_save=None):
+    """
+    Allows to associate a cell type to a prediction (a prediction being a 2d array of n_cells * n_classes)
+    Args:
+        cell_type_preds_dict: key: code of the cell type, value 2d array of n_cells * n_classes representing
+        the predictions to associate to this cell type (for cell predicted as so)
+        default_cell_type_pred: 2d array of n_cells * n_classes
+        Same length as cell_type_preds_dict
+        cell_type_config_file:
+        filename_to_save:
+        cell_type_to_skip_if_conflict: int representing the cell type code, if given and there is a conflict between
+        two classifier including this cell type, then the other one in conflict will be chosen. If a conflict with more
+        than 2 cell types, then the default one is chosen.
+
+    Returns:
+
+    """
+    n_cells = len(default_cell_type_pred)
+    cell_type_codes = list(cell_type_preds_dict.keys())
+
+    if cell_type_config_file is not None:
+        cell_type_from_code_dict, cell_type_to_code_dict, multi_class_arg = \
+            read_cell_type_categories_yaml_file(yaml_file=cell_type_config_file)
+        # we display info about the code:
+        print(f"fusion_cell_type_predictions_by_type() on {n_cells} cells")
+        for code in cell_type_codes:
+            print(f"Cell type for code {code}: {cell_type_from_code_dict[code]}")
+        print(" ")
+    else:
+        cell_type_from_code_dict = None
+
+    # key is the cell type code, value is a list of cells (int)
+    cells_by_type = dict()
+    # key is the cell type code, value is a list of array of lentgh n_cell_types
+    cells_predictions_by_type = dict()
+    for cell_type_code in cell_type_codes:
+        cells_by_type[cell_type_code] = []
+        cells_predictions_by_type[cell_type_code] = []
+
+    for cell_type_code, predictions in cell_type_preds_dict.items():
+        for cell in range(n_cells):
+            if np.argmax(predictions[cell]) == cell_type_code:
+                cells_by_type[cell_type_code].append(cell)
+                cells_predictions_by_type[cell_type_code].append(predictions[cell])
+
+    # now we look at how much intersects between cell_type
+    # without intersects
+
+    # final predictions
+    fusion_predictions = np.zeros((n_cells, len(cell_type_codes)))
+
+    n_cells_that_interesect = 0
+    n_cells_non_predicted = 0
+    interesect_association = []
+    if cell_type_from_code_dict is not None:
+        print(f"Cells with multiple cell type (-> default that will the final one):")
+    for cell in range(n_cells):
+        cell_type_code_associated = []
+        cell_predictions = None
+        for cell_type_code, cells in cells_by_type.items():
+            if cell in cells:
+                cell_type_code_associated.append(cell_type_code)
+                index_cell = np.where(np.array(cells) == cell)[0][0]
+                cell_predictions = cells_predictions_by_type[cell_type_code][index_cell]
+        if len(cell_type_code_associated) == 1:
+            # cell_type_code = cell_type_code_associated[0]
+            fusion_predictions[cell] = cell_predictions
+        else:
+            use_default = True
+            if len(cell_type_code_associated) == 0:
+                n_cells_non_predicted += 1
+            else:
+                if cell_type_from_code_dict is not None:
+                    new_cell_type_code = np.argmax(default_cell_type_pred[cell])
+                    interesect_association.append(tuple([cell_type_from_code_dict[c]
+                                                         for c in cell_type_code_associated]))
+                    if (cell_type_to_skip_if_conflict is not None) and len(cell_type_code_associated) == 2:
+                        if cell_type_to_skip_if_conflict in cell_type_code_associated:
+                            use_default = False
+                            if cell_type_to_skip_if_conflict == cell_type_code_associated[0]:
+                                new_cell_type_code = cell_type_code_associated[1]
+                            else:
+                                new_cell_type_code = cell_type_code_associated[0]
+                            cells = cells_by_type[new_cell_type_code]
+                            index_cell = np.where(np.array(cells) == cell)[0][0]
+                            fusion_predictions[cell] = cells_predictions_by_type[new_cell_type_code][index_cell]
+                    print(f"Cell {cell}: {[cell_type_from_code_dict[c] for c in cell_type_code_associated]} -> "
+                          f"{cell_type_from_code_dict[new_cell_type_code]}")
+                n_cells_that_interesect += 1
+            if use_default:
+                fusion_predictions[cell] = default_cell_type_pred[cell]
+
+    if cell_type_config_file is not None:
+        print(f"N cells without predictions on first round: {n_cells_non_predicted}")
+        print(f"N cells with more than on cell type on first round: {n_cells_that_interesect}")
+        print(f"Associations that intersect: {Counter(interesect_association)}")
+
+    if filename_to_save is not None:
+        if not filename_to_save.endswith(".npy"):
+            filename_to_save = filename_to_save + ".npy"
+        np.save(filename_to_save, fusion_predictions)
+
+    print(" ")
+
+    if cell_type_from_code_dict is not None:
+        # now we want to look at how many cells were wrongly classified in each pred
+        for main_cell_type_code, predictions in cell_type_preds_dict.items():
+            wrong_classifications_by_cell = np.zeros(len(cell_type_preds_dict), dtype="int16")
+            for cell in range(n_cells):
+                if np.argmax(fusion_predictions[cell]) != np.argmax(predictions[cell]):
+                    wrong_cell_type = np.argmax(predictions[cell])
+                    wrong_classifications_by_cell[wrong_cell_type] += 1
+            print(f"Wrong classifications for predictions associated "
+                  f"to {cell_type_from_code_dict[main_cell_type_code]} :")
+            for cell_type_code in range(len(cell_type_preds_dict)):
+                n_wrong = wrong_classifications_by_cell[cell_type_code]
+                cell_type_name = cell_type_from_code_dict[cell_type_code]
+                print(f"{n_wrong} wrongs {cell_type_name}")
+
+        print(" ")
+        # we display to count by cell type in the fusion predictions
+        cell_count_by_type = np.zeros(fusion_predictions.shape[1], dtype="int16")
+        for cell in np.arange(len(fusion_predictions)):
+            cell_type_code = np.argmax(fusion_predictions[cell])
+            cell_count_by_type[cell_type_code] += 1
+        print(" ")
+        print(f"Number of cells for each cell type after fusion:")
+        for code in np.arange(len(cell_type_from_code_dict)):
+            print(f"{cell_count_by_type[code]} {cell_type_from_code_dict[code]}")
+        print(" ")
+
+    return fusion_predictions
+
+
 
 
 def fusion_cell_type_predictions(cell_type_pred_1, cell_type_pred_2,
@@ -404,7 +543,7 @@ class CinacPredictor:
                                                                    all_pixels=all_pixels,
                                                                    n_segments_to_use_for_prediction=
                                                                    n_segments_to_use_for_prediction,
-                                                                   verbose=self.verbose)
+                                                                   verbose=int(time_verbose))
                         # print(f"predictions in predict() {len(predictions)} {len(predictions[0])} {predictions}")
                         if len(predictions) > 1:
                             if len(predictions[0]) == 1:
@@ -458,7 +597,7 @@ class CinacPredictor:
             if len(network_identifiers) == 0:
                 network_identifier = ""
             else:
-                network_identifier = "_" + "_".join(network_identifiers)
+                network_identifier = "_".join(network_identifiers)
             if results_path is not None:
                 if create_dir_for_results:
                     time_str = datetime.now().strftime("%Y_%m_%d.%H-%M-%S")
@@ -470,7 +609,7 @@ class CinacPredictor:
                 predictions_str = "predictions"
                 if cell_type_classifier_mode:
                     predictions_str = "cell_type_predictions"
-                file_name = f"{cinac_recording.identifier}_{predictions_str}{network_identifier}"
+                file_name = f"{cinac_recording.identifier}_{predictions_str}_{network_identifier}"
                 for output_file_format in output_file_formats:
                     if "mat" in output_file_format:
                         sio.savemat(os.path.join(new_dir_results, file_name + ".mat"),
@@ -487,6 +626,24 @@ def predict_cell_type_from_model(cinac_recording, cell, model,
                                  n_frames, n_segments_to_use_for_prediction=2,
                                  overlap_value=0, pixels_around=0, all_pixels=False,
                                  use_data_augmentation=False, buffer=None, verbose=0):
+    """
+
+    Args:
+        cinac_recording:
+        cell:
+        model:
+        n_frames:
+        n_segments_to_use_for_prediction:
+        overlap_value:
+        pixels_around:
+        all_pixels:
+        use_data_augmentation:
+        buffer:
+        verbose:
+
+    Returns:
+
+    """
     # if n_frames is None, then the movie need to have been loaded
     # start_time = time.time()
     # multi_inputs = (model.layers[0].output_shape == model.layers[1].output_shape)
